@@ -18,11 +18,10 @@ This module is a thin layer of abstraction around the library.
 It exposes all core functionality.
 """
 
-import capabilities
-import operations
-import transport
-
+from ncclient import operations
+from ncclient import transport
 import logging
+import functools
 
 from ncclient.xml_ import *
 
@@ -31,15 +30,18 @@ logger = logging.getLogger('ncclient.manager')
 OPERATIONS = {
     "get": operations.Get,
     "get_config": operations.GetConfig,
+    "get_schema": operations.GetSchema,
     "dispatch": operations.Dispatch,
     "edit_config": operations.EditConfig,
     "copy_config": operations.CopyConfig,
     "validate": operations.Validate,
     "commit": operations.Commit,
     "discard_changes": operations.DiscardChanges,
+    "cancel_commit": operations.CancelCommit,
     "delete_config": operations.DeleteConfig,
     "lock": operations.Lock,
     "unlock": operations.Unlock,
+    "create_subscription": operations.CreateSubscription,
     "close_session": operations.CloseSession,
     "kill_session": operations.KillSession,
     "poweroff_machine": operations.PoweroffMachine,
@@ -53,8 +55,6 @@ subclasses. It is used to lookup operations, e.g. `get_config` is mapped to
 operations to the :class:`Manager` API.
 """
 
-VENDOR_OPERATIONS = {}
-
 
 def make_device_handler(device_params):
     """
@@ -68,6 +68,10 @@ def make_device_handler(device_params):
     if device_params is None:
         device_params = {}
 
+    handler = device_params.get('handler', None)
+    if handler:
+        return handler(device_params)
+
     device_name = device_params.get("name", "default")
     # Attempt to import device handler class. All device handlers are
     # in a module called "ncclient.devices.<devicename>" and in a class named
@@ -81,102 +85,94 @@ def make_device_handler(device_params):
     return handler_obj
 
 
+def _extract_device_params(kwds):
+    device_params = kwds.pop("device_params", None)
+
+    return device_params
+
+def _extract_manager_params(kwds):
+    manager_params = kwds.pop("manager_params", {})
+
+    # To maintain backward compatibility
+    if 'timeout' not in manager_params and 'timeout' in kwds:
+        manager_params['timeout'] = kwds['timeout']
+    return manager_params
+
+
 def connect_ssh(*args, **kwds):
     """
     Initialize a :class:`Manager` over the SSH transport.
     For documentation of arguments see :meth:`ncclient.transport.SSHSession.connect`.
 
     The underlying :class:`ncclient.transport.SSHSession` is created with
-        :data:`CAPABILITIES`. It is first instructed to
-        :meth:`~ncclient.transport.SSHSession.load_known_hosts` and then
-        all the provided arguments are passed directly to its implementation
-        of :meth:`~ncclient.transport.SSHSession.connect`.
+    :data:`CAPABILITIES`. It is first instructed to
+    :meth:`~ncclient.transport.SSHSession.load_known_hosts` and then
+    all the provided arguments are passed directly to its implementation
+    of :meth:`~ncclient.transport.SSHSession.connect`.
 
-    To invoke advanced vendor related operation add device_params =
-        {'name':'<vendor_alias>'} in connection paramerers. For the time,
-        'junos' and 'nexus' are supported for Juniper and Cisco Nexus respectively.
+    To customize the :class:`Manager`, add a `manager_params` dictionary in connection
+    parameters (e.g. `manager_params={'timeout': 60}` for a bigger RPC timeout parameter)
+
+    To invoke advanced vendor related operation add
+    `device_params={'name': '<vendor_alias>'}` in connection parameters. For the time,
+    'junos' and 'nexus' are supported for Juniper and Cisco Nexus respectively.
+
+    A custom device handler can be provided with
+    `device_params={'handler':<handler class>}` in connection parameters.
     """
-    # Extract device parameter dict, if it was passed into this function. Need to
-    # remove it from kwds, since the session.connect() doesn't like extra stuff in
-    # there.
-    if "device_params" in kwds:
-        device_params = kwds["device_params"]
-        del kwds["device_params"]
-    else:
-        device_params = None
+    # Extract device parameter and manager parameter dictionaries, if they were passed into this function.
+    # Remove them from kwds (which should keep only session.connect() parameters).
+    device_params = _extract_device_params(kwds)
+    manager_params = _extract_manager_params(kwds)
 
     device_handler = make_device_handler(device_params)
     device_handler.add_additional_ssh_connect_params(kwds)
-    global VENDOR_OPERATIONS
-    VENDOR_OPERATIONS.update(device_handler.add_additional_operations())
     session = transport.SSHSession(device_handler)
-    session.load_known_hosts()
+    if "hostkey_verify" not in kwds or kwds["hostkey_verify"]:
+        session.load_known_hosts()
 
     try:
        session.connect(*args, **kwds)
     except Exception as ex:
         if session.transport:
             session.close()
-        raise ex
-    return Manager(session, device_handler, **kwds)
+        raise
+    return Manager(session, device_handler, **manager_params)
+
 
 def connect_ioproc(*args, **kwds):
-    if "device_params" in kwds:
-        device_params = kwds["device_params"]
-        del kwds["device_params"]
+    device_params = _extract_device_params(kwds)
+    manager_params = _extract_manager_params(kwds)
+
+    if device_params:
         import_string = 'ncclient.transport.third_party.'
         import_string += device_params['name'] + '.ioproc'
         third_party_import = __import__(import_string, fromlist=['IOProc'])
-    else:
-        device_params = None
 
     device_handler = make_device_handler(device_params)
 
-    global VENDOR_OPERATIONS
-    VENDOR_OPERATIONS.update(device_handler.add_additional_operations())
     session = third_party_import.IOProc(device_handler)
     session.connect()
 
-    return Manager(session, device_handler, **kwds)
+    return Manager(session, device_handler, **manager_params)
 
 
 def connect(*args, **kwds):
     if "host" in kwds:
         host = kwds["host"]
-        if host != 'localhost':
-            return connect_ssh(*args, **kwds)
-        else:
+        device_params = kwds.get('device_params', {})
+        if host == 'localhost' and device_params.get('name') == 'junos' \
+                and device_params.get('local'):
             return connect_ioproc(*args, **kwds)
-
-class OpExecutor(type):
-
-    def __new__(cls, name, bases, attrs):
-        def make_wrapper(op_cls):
-            def wrapper(self, *args, **kwds):
-                return self.execute(op_cls, *args, **kwds)
-            wrapper.func_doc = op_cls.request.func_doc
-            return wrapper
-        for op_name, op_cls in OPERATIONS.iteritems():
-            attrs[op_name] = make_wrapper(op_cls)
-        return super(OpExecutor, cls).__new__(cls, name, bases, attrs)
-
-    def __call__(cls, *args, **kwargs):
-        def make_wrapper(op_cls):
-            def wrapper(self, *args, **kwds):
-                return self.execute(op_cls, *args, **kwds)
-            wrapper.func_doc = op_cls.request.func_doc
-            return wrapper
-        if VENDOR_OPERATIONS:
-            for op_name, op_cls in VENDOR_OPERATIONS.iteritems():
-                setattr(cls, op_name, make_wrapper(op_cls))
-        return super(OpExecutor, cls).__call__(*args, **kwargs)
+        else:
+            return connect_ssh(*args, **kwds)
 
 
 class Manager(object):
 
     """
     For details on the expected behavior of the operations and their
-        parameters refer to :rfc:`4741`.
+        parameters refer to :rfc:`6241`.
 
     Manager instances are also context managers so you can use it like this::
 
@@ -192,14 +188,22 @@ class Manager(object):
             m.close_session()
     """
 
-    __metaclass__ = OpExecutor
+   # __metaclass__ = OpExecutor
 
-    def __init__(self, session, device_handler, timeout=30, *args, **kwargs):
+
+    HUGE_TREE_DEFAULT = False
+    """Default for `huge_tree` support for XML parsing of RPC replies (defaults to False)"""
+
+    def __init__(self, session, device_handler, timeout=30):
         self._session = session
         self._async_mode = False
         self._timeout = timeout
         self._raise_mode = operations.RaiseMode.ALL
+        self._huge_tree = self.HUGE_TREE_DEFAULT
         self._device_handler = device_handler
+        self._vendor_operations = {}
+        if device_handler:
+            self._vendor_operations.update(device_handler.add_additional_operations())
 
     def __enter__(self):
         return self
@@ -221,9 +225,10 @@ class Manager(object):
     def execute(self, cls, *args, **kwds):
         return cls(self._session,
                    device_handler=self._device_handler,
-                   async=self._async_mode,
+                   async_mode=self._async_mode,
                    timeout=self._timeout,
-                   raise_mode=self._raise_mode).request(*args, **kwds)
+                   raise_mode=self._raise_mode,
+                   huge_tree=self._huge_tree).request(*args, **kwds)
 
     def locked(self, target):
         """Returns a context manager for a lock on a datastore, where
@@ -249,16 +254,39 @@ class Manager(object):
         raise NotImplementedError
 
     def __getattr__(self, method):
-        """Parse args/kwargs correctly in order to build XML element"""
-        def _missing(*args, **kwargs):
-            m = method.replace('_', '-')
-            root = new_ele(m)
-            if args:
-                for arg in args:
-                    sub_ele(root, arg)
-            r = self.rpc(root)
-            return r
-        return _missing
+        if method in self._vendor_operations:
+            return functools.partial(self.execute, self._vendor_operations[method])
+        elif method in OPERATIONS:
+            return functools.partial(self.execute, OPERATIONS[method])
+        else:
+            """Parse args/kwargs correctly in order to build XML element"""
+            def _missing(*args, **kwargs):
+                m = method.replace('_', '-')
+                root = new_ele(m)
+                if args:
+                    for arg in args:
+                        sub_ele(root, arg)
+                r = self.rpc(root)
+                return r
+            return _missing
+
+    def take_notification(self, block=True, timeout=None):
+        """Attempt to retrieve one notification from the queue of received
+        notifications.
+
+        If block is True, the call will wait until a notification is
+        received.
+
+        If timeout is a number greater than 0, the call will wait that
+        many seconds to receive a notification before timing out.
+
+        If there is no notification available when block is False or
+        when the timeout has elapse, None will be returned.
+
+        Otherwise a :class:`~ncclient.operations.notify.Notification`
+        object will be returned.
+        """
+        return self._session.take_notification(block, timeout)
 
     @property
     def client_capabilities(self):
@@ -304,3 +332,13 @@ class Manager(object):
     exceptions. Valid values are the constants defined in
     :class:`~ncclient.operations.RaiseMode`.
     The default value is :attr:`~ncclient.operations.RaiseMode.ALL`."""
+
+    @property
+    def huge_tree(self):
+        """Whether `huge_tree` support for XML parsing of RPC replies is enabled (default=False)
+        The default value is configurable through :attr:`~ncclient.manager.Manager.HUGE_TREE_DEFAULT`"""
+        return self._huge_tree
+
+    @huge_tree.setter
+    def huge_tree(self, x):
+        self._huge_tree = x
